/* Talika panel applet: Switch between open windows using icons.
 * (C) 2009-2010 Sinew Software Systems
	 *
 * Authors:  Vinod Kumar Dahiya
 *           Hemant Kumar
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
	 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <gtk/gtk.h>
#include <gio/gio.h>
#include <gio/gdesktopappinfo.h>

#include "task.h"
#include "launcher.h"
#include "tasklist.h"
#include "tooltips.h"
#include "button-widget.h"

static void 
talika_launcher_update_visible_state (TalikaTask * launcher);

/*########################## Utility functions ###############################*/

static gchar *
get_process_name_from_pid (gint pid)
{
	if(pid==0)
		return NULL;

	gchar *name=NULL;
	gchar *file;
	file = g_strdup_printf ("/proc/%d/cmdline",pid);

	gchar *contents;
	gsize len;
	if(g_file_get_contents(file,&contents,&len,NULL))
	{
		//every argument of commandline is terminated by null
		//so arguments other than first already discarded
		name = contents;
	}
	
	g_free(file);

	//g_print (" pid : %d , process : %s\n",pid,name);
	
	return name;
}

static GAppInfo*
talika_launcher_get_app_info_from_id (const gchar *id_name)
{
	gchar *filename;

	//get desktop file from class name first
	gchar *temp = g_strdup_printf ("%s.desktop",id_name);

	GAppInfo *app_info = (GAppInfo *)g_desktop_app_info_new (temp);
	if (!app_info)
	{	
		filename = g_utf8_strdown (temp,-1);

		app_info = (GAppInfo *)g_desktop_app_info_new (filename);
		g_free(filename);
	}

	g_free (temp);

	return app_info;
}

static void 
do_free_app_info (gpointer data, 
                    gpointer user_data)
{
	GAppInfo *app_info = (GAppInfo *) data;
	g_object_unref (app_info);
}

static GAppInfo*
talika_launcher_get_app_info_from_process_name (const gchar *process_name)
{
	gchar *pname = g_path_get_basename (process_name);

	GAppInfo *app = NULL;
	
	GList *app_list;
	app_list = g_app_info_get_all ();

	GList *l;
	for (l = app_list; l; l = l->next)
	{
		GAppInfo *app_info = l->data;

		const gchar *tmp = g_app_info_get_executable (app_info);

		//user case crash pining chrome, testing
		if(!tmp)
			continue;
		
		gchar **splits = g_strsplit (tmp," ",2);

		gchar *ename = g_path_get_basename (splits[0]);
		g_strfreev (splits);

		if(g_ascii_strcasecmp (ename,pname) == 0)
		{			
			app = g_app_info_dup (app_info);
			g_free (ename);
			break;
		}

		g_free (ename);
	}

	g_list_foreach (app_list,do_free_app_info,NULL);
	g_list_free (app_list);
	g_free (pname);
	return app;
}

static GAppInfo*
talika_launcher_get_app_info_from_name (const gchar *name)
{
	gchar *name1 = g_utf8_casefold (name,-1);
	GAppInfo *app = NULL;
	
	GList *app_list;
	app_list = g_app_info_get_all ();

	GList *l;
	for (l = app_list; l; l = l->next)
	{
		GAppInfo *app_info = l->data;

		const gchar *tmp = g_app_info_get_name (app_info);
		gchar *name2 = g_utf8_casefold (tmp,-1);
		
		gint len = strlen(name1);

		if(strncmp (name1,name2,len) == 0)
		{			
			app = g_app_info_dup (app_info);
			g_free (name2);
			break;
		}

		g_free (name2);
	}

	g_list_foreach (app_list,do_free_app_info,NULL);
	g_list_free (app_list);
	g_free (name1);
	return app;
}

static GAppInfo*
talika_launcher_get_app_info (const gchar *name, gint pid)
{	
	//wine apps are not supported, return immediatly
	//if( g_ascii_strcasecmp (name,"Wine")==0)
	//	return NULL;
	
	GAppInfo *app_info =NULL;
	gchar *process_name = get_process_name_from_pid(pid);

	//try to match class name with desktop id
	app_info = talika_launcher_get_app_info_from_id (name);

	//try to match process name with desktop id
	if(!app_info)
	{
		if(process_name)
		{
			app_info = talika_launcher_get_app_info_from_id (process_name);
		}
	}

	//try to match process name with process names
	if(!app_info)
	{
		if(process_name)
		{
			app_info = 
				talika_launcher_get_app_info_from_process_name (process_name);
		}
	}

	//try to match name with names
	if(!app_info)
	{
			app_info = 
				talika_launcher_get_app_info_from_name (name);
	}
	
	if(process_name)
		g_free (process_name);
	
	return app_info;
}

static GdkPixbuf*
talika_get_icon_from_app_info (const GAppInfo* app_info, guint size)
{
	GdkPixbuf *pixbuf=NULL;
	/*pixbuf = gdk_pixbuf_new_from_file_at_size ("/usr/share/pixmaps/apple-red.png",
		                                           size,
		                                           size,
		                                           NULL);
	return pixbuf;*/
	
	if(app_info)
	{
		GIcon *icon = g_app_info_get_icon(G_APP_INFO(app_info));

		if(G_IS_FILE_ICON(icon))
		{
			GFile *icon_file = g_file_icon_get_file (G_FILE_ICON (icon));
			gchar *icon_name = g_file_get_path (icon_file);
			pixbuf = gdk_pixbuf_new_from_file_at_size (icon_name,
			                                           size,
			                                           size,
			                                           NULL);
			g_free (icon_name);
			//g_object_unref (icon_file);
		}
		else if(G_IS_THEMED_ICON(icon))
		{
			const gchar**icon_names =
				(const gchar**)g_themed_icon_get_names(G_THEMED_ICON(icon));

			GtkIconTheme *icon_theme = gtk_icon_theme_get_default ();
			GtkIconInfo *icon_info =
				gtk_icon_theme_choose_icon (icon_theme,icon_names,size,0);

			if(icon_info)
			{
				const gchar *icon_name = gtk_icon_info_get_filename (icon_info);
				
				pixbuf = gdk_pixbuf_new_from_file_at_size (icon_name,
			                                           size,
			                                           size,
			                                           NULL);

				gtk_icon_info_free (icon_info);
			}
		}
	}

	return pixbuf;
}

GdkPixbuf*
talika_get_icon_from_desktop_file(TalikaTaskList *tasklist, const gchar *desktop_id, guint size)
{
	if (!desktop_id)
		return NULL;
	
	GdkPixbuf *pixbuf=NULL;

	//try to find in cache
	pixbuf = talika_icon_cache_get_icon (tasklist->icon_cache,
	                                                          desktop_id);
	
	if(!pixbuf)
	{
		GAppInfo *app_info = talika_launcher_get_app_info_from_id (desktop_id);

		if(app_info)
		{
			pixbuf = talika_get_icon_from_app_info (app_info, size);
			g_object_unref (app_info);
			//g_print ("from appinfo\n");

			talika_icon_cache_add (tasklist->icon_cache,desktop_id,pixbuf);
		}
	}

	return pixbuf;
}

gchar*
talika_get_spec_from_class_pid (const gchar *name, gint pid)
{
	gchar *desktop_id = NULL;
	gchar *spec = NULL;
	
	GAppInfo *app_info = talika_launcher_get_app_info(name, pid);
	
	if(app_info)
	{
		const gchar *temp = g_app_info_get_id (app_info);
		gint len = strlen(temp);
		desktop_id = g_strndup(temp,len-8);
		
		spec = g_strdup_printf ("%s:%s",desktop_id,name);
		g_print ("%s\n",spec);
		g_print ("\n");

		g_object_unref (app_info);
		g_free (desktop_id);
	}

	return spec;
}

gchar*
talika_determine_desktop_id_for_window (WnckWindow *window, gchar *class_name)
{
	gchar *desktop_id = NULL;
	
	gint pid = wnck_window_get_pid (window);
	GAppInfo *app_info = talika_launcher_get_app_info(class_name, pid);

	if(app_info)
	{
		const gchar *temp = g_app_info_get_id (app_info);
		gint len = strlen(temp);
		desktop_id = g_strndup(temp,len-8);
	
		g_object_unref (app_info);
	}

	return desktop_id;
}

static gchar*
get_last_part_of_title (const gchar *title)
{
	g_assert (title !=NULL);
	
	gchar *ret;
	gchar **tokens = g_strsplit (title,"-",-1);

	//get last part
	gint i=0;
	while (tokens[i+1]!=NULL)
		i++;

	ret = g_strdup (tokens[i]);
	g_strstrip (ret);

	g_strfreev (tokens);

	return ret;
}

gchar*
talika_determine_class_name_for_window (WnckWindow *window)
{
	gchar *name;
	WnckClassGroup *c_group = 
				wnck_window_get_class_group (window);
	const gchar *class_name = wnck_class_group_get_res_class (c_group);

	//g_print("%s\n",class_name);

	//make some assumption hereabout ooo and wine apps
	if (g_ascii_strncasecmp (class_name,"OpenOffice.org",14)==0)
	{
		//openoffice app
		name = get_last_part_of_title (wnck_window_get_name (window));
		g_strdelimit (name, " ", '-');
	}
	else if (g_ascii_strncasecmp (class_name,"Wine",4)==0)
	{
		//if title is picasa
		if (g_ascii_strncasecmp (wnck_window_get_name (window),"Picasa",6)==0)
		{
			name = g_strdup ("Picasa");
		}
		else
		{
			//wine app
			gchar *temp;
			temp = get_last_part_of_title (wnck_window_get_name (window));
			g_strdelimit (temp, " ", '-');
			name = g_strdup_printf ("wine-%s",temp);
			g_free (temp);
		}
	}
	else if (g_ascii_strncasecmp (class_name,"Googleearth",11)==0)
	{
		name = g_strdup ("Google Earth");
	}
	else
	{
		name = g_strdup (class_name);
	}

	return name;
}

/*############################### Class cache ################################*/
static void
talika_class_cache_value_free(gpointer value)
{
	TalikaClass *class_obj = value;
	g_free (class_obj->class_name);
	g_free (class_obj->desktop_id);
	g_free (class_obj);
	class_obj = NULL;
}

TalikaClassCache* 
talika_class_cache_new(void)
{
	TalikaClassCache *cache;
	cache = g_new0(TalikaClassCache,1);
	cache->hash = g_hash_table_new_full (NULL,NULL,NULL,
	                                     talika_class_cache_value_free);
	cache->id_hash = g_hash_table_new_full (g_str_hash,g_str_equal,g_free,g_free);
	return cache;
}

void
talika_class_cache_free(TalikaClassCache *cache)
{
	if(!cache)
		return;
	
	g_hash_table_unref (cache->hash);
	g_hash_table_unref (cache->id_hash);
	g_free (cache);
}

void
talika_class_cache_update_window (TalikaClassCache *cache,
                                          WnckWindow *window)
{
	if(!cache)
	{
		return;
	}
	
	
	TalikaClass *class_obj=NULL;

	//determine class for window
	class_obj = g_new0(TalikaClass,1);

	class_obj->class_name = talika_determine_class_name_for_window (window);

	//find desktop id in hash first
	gchar *id = g_hash_table_lookup (cache->id_hash,class_obj->class_name);
	if(id)
	{
		class_obj->desktop_id = g_strdup (id);
		//g_print ("id from cache\n");
	}
	else
	{
		//g_print ("id from other\n");
		
		class_obj->desktop_id = 
			talika_determine_desktop_id_for_window (window,class_obj->class_name);

		if(class_obj->desktop_id)
			g_hash_table_replace (cache->id_hash,g_strdup(class_obj->class_name),
			                      g_strdup(class_obj->desktop_id));
	}

	g_hash_table_replace (cache->hash,window,class_obj);

	g_print ("class_name : %s\n",class_obj->class_name);
	if (class_obj->desktop_id)
		g_print ("desktop_id : %s\n", class_obj->desktop_id);
}

TalikaClass*
talika_class_cache_get_class (TalikaClassCache *cache,
                                      WnckWindow *window)
{
	g_assert (cache !=NULL);

	return g_hash_table_lookup (cache->hash,window);
}

void
talika_class_cache_free_window (TalikaClassCache *cache,
                                        WnckWindow *window)
{
	if(!cache)
		return;
	
	g_hash_table_remove (cache->hash,window);
}
/*############################### Icon cache ################################*/

TalikaIconCache* 
talika_icon_cache_new(void)
{
	TalikaIconCache *cache;
	cache = g_new0(TalikaIconCache,1);
	cache->hash = g_hash_table_new_full (g_str_hash,g_str_equal,g_free,g_object_unref);
	return cache;
}

void
talika_icon_cache_free(TalikaIconCache *cache)
{
	if(!cache)
		return;
	
	g_hash_table_unref (cache->hash);
	g_free (cache);
}

void
talika_icon_cache_add (TalikaIconCache *cache,
                       const gchar *desktop_id,
                       GdkPixbuf *icon)
{
	if(!cache)
	{
		return;
	}
	
	g_hash_table_replace (cache->hash,g_strdup(desktop_id),g_object_ref(icon));

}

GdkPixbuf*
talika_icon_cache_get_icon(TalikaIconCache *cache,
                                      const gchar *desktop_id)
{
	g_assert (cache !=NULL);
	GdkPixbuf *icon=NULL;
	
	icon = g_hash_table_lookup (cache->hash,desktop_id);
	if(icon)
		g_object_ref(icon);
	
	return icon;
}


/*############################### Launcher ###################################*/

GdkPixbuf* 
talika_launcher_get_icon (TalikaTask *launcher)
{
	GdkPixbuf *pixbuf = NULL;
	guint size = launcher->tasklist->panel_size - launcher->tasklist->icon_size_offset;
	
	if (launcher->icon_path)
	{
		//g_print (launcher->icon_path);
		pixbuf = gdk_pixbuf_new_from_file_at_size (launcher->icon_path,
		                                           size,
		                                           size,
		                                           NULL);
	}
	else
	{
		//try to find in cache
		pixbuf = talika_get_icon_from_desktop_file (launcher->tasklist,launcher->desktop_id,size);
	}

	return pixbuf;
}

static void
talika_launcher_position_menu (GtkMenu * menu,
                           int *x,
                           int *y, gboolean * push_in, gpointer user_data)
{
	TalikaTask *task = (TalikaTask *) user_data;
	GtkWidget *widget = task->button;

	GtkRequisition requisition;
	GdkScreen *screen;
	int menu_x = 0;
	int menu_y = 0;
	int pointer_x;
	int pointer_y;

	screen = gtk_widget_get_screen (widget);

	gtk_widget_size_request (GTK_WIDGET (menu), &requisition);

	gdk_window_get_origin (widget->window, &menu_x, &menu_y);
	gtk_widget_get_pointer (widget, &pointer_x, &pointer_y);

	//works for buttons only, not event box
	menu_x += widget->allocation.x;
	menu_y += widget->allocation.y;

	if (task->tasklist->panel_orient == PANEL_APPLET_ORIENT_UP ||
	    task->tasklist->panel_orient == PANEL_APPLET_ORIENT_DOWN)
	{
		if (gtk_widget_get_direction (GTK_WIDGET (menu)) !=
		    GTK_TEXT_DIR_RTL)
		{
			if (pointer_x < widget->allocation.width &&
			    requisition.width < pointer_x)
				menu_x += MIN (pointer_x,
				               widget->allocation.width -
				               requisition.width);
		}
		else
		{
			menu_x +=
				widget->allocation.width - requisition.width;
			if (pointer_x > 0
			    && pointer_x < widget->allocation.width
			    && pointer_x <
			    widget->allocation.width - requisition.width)
			{
				menu_x -=
					MIN (widget->allocation.width -
					     pointer_x,
					     widget->allocation.width -
					     requisition.width);
			}
		}
		menu_x = MIN (menu_x,
		              gdk_screen_get_width (screen) -
		              requisition.width);

		if (menu_y > gdk_screen_get_height (screen) / 2)
			menu_y -= requisition.height;
		else
			menu_y += widget->allocation.height;
	}
	else
	{
		if (pointer_y < widget->allocation.height &&
		    requisition.height < pointer_y)
			menu_y +=
			MIN (pointer_y,
			     widget->allocation.height -
			     requisition.height);
		menu_y = MIN (menu_y,
		              gdk_screen_get_height (screen) -
		              requisition.height);

		if (menu_x > gdk_screen_get_width (screen) / 2)
			menu_x -= requisition.width;
		else
			menu_x += widget->allocation.width;

	}

	*x = menu_x;
	*y = menu_y;
	*push_in = TRUE;
}

void 
talika_launcher_launch (GtkWidget *menu_item, TalikaTask *launcher)
{
	/* Did we really want to change the state of the togglebutton? */
	if (launcher->really_toggling)
		return;

	/* Undo the toggle */
	launcher->really_toggling = TRUE;
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (launcher->button),FALSE);
	launcher->really_toggling = FALSE;
	
	//gtk version older than 2.14 doesn't have GdkAppLaunchContext, so
	//here is a fake replacement
#if GTK_MINOR_VERSION < 14
	GAppLaunchContext *context;
	context = g_app_launch_context_new ();
#else	
	GdkAppLaunchContext *context;
	context = gdk_app_launch_context_new ();
	gdk_app_launch_context_set_screen (context, gtk_widget_get_screen (menu_item));
	gdk_app_launch_context_set_timestamp (context, gtk_get_current_event_time());
#endif
	
	if (g_app_info_launch(G_APP_INFO(launcher->app_info),
	                      NULL,G_APP_LAUNCH_CONTEXT(context),NULL))
	{
		//sucessfull
		if(launcher->button)
			xstuff_zoom_animate (launcher->button,
			                     talika_task_get_icon(launcher),
			                     launcher->tasklist->panel_orient,
			                     NULL);
	}
	else
	{
		//report error
	}

    g_object_unref (context);
}

void 
talika_launcher_launch_uris (GtkWidget *menu_item, 
                             TalikaTask *launcher,
                             GList *uris)
{
	//gtk version older than 2.14 doesn't have GdkAppLaunchContext, so
	//here is a fake replacement
#if GTK_MINOR_VERSION < 14
	GAppLaunchContext *context;
	context = g_app_launch_context_new ();
#else	
	GdkAppLaunchContext *context;
	context = gdk_app_launch_context_new ();
	gdk_app_launch_context_set_screen (context, gtk_widget_get_screen (menu_item));
	gdk_app_launch_context_set_timestamp (context, gtk_get_current_event_time());
#endif
	
	if (g_app_info_launch_uris (G_APP_INFO(launcher->app_info),
	                      uris,G_APP_LAUNCH_CONTEXT(context),NULL))
	{
		//sucessfull
		if(launcher->button)
			xstuff_zoom_animate (launcher->button,
			                     talika_task_get_icon(launcher),
			                     launcher->tasklist->panel_orient,
			                     NULL);
	}
	else
	{
		//report error
	}

    g_object_unref (context);
}

void 
talika_launcher_unpin (GtkWidget *menu_item, TalikaTask *launcher)
{
	talika_tasklist_remove_launcher (launcher);
}

static void 
talika_launcher_attach_to_task (GtkWidget *menu_item, TalikaTask *launcher)
{
	const gchar *class_name = 
		gtk_label_get_text (GTK_LABEL(GTK_BIN (menu_item)->child));

	g_free(launcher->class_name);
	launcher->class_name = g_strdup (class_name);
	talika_tasklist_save_launchers (launcher->tasklist);

	//update visible tasks now
	talika_tasklist_update_lists (launcher->tasklist);
	//gtk_widget_queue_resize (GTK_WIDGET (launcher->tasklist->box));
}

static void
update_preview_cb (GtkFileChooser *file_chooser, gpointer data)
{
  GtkWidget *preview;
  char *filename;
  GdkPixbuf *pixbuf;
  gboolean have_preview;

  preview = GTK_WIDGET (data);
  filename = gtk_file_chooser_get_preview_filename (file_chooser);

  pixbuf = gdk_pixbuf_new_from_file_at_size (filename, 128, 128, NULL);
  have_preview = (pixbuf != NULL);
  g_free (filename);

  gtk_image_set_from_pixbuf (GTK_IMAGE (preview), pixbuf);
  if (pixbuf)
    g_object_unref (pixbuf);

  gtk_file_chooser_set_preview_widget_active (file_chooser, have_preview);
}

void 
talika_launcher_change_icon (GtkWidget *menu_item, TalikaTask *launcher)
{
	GtkWidget * dlg;
	GtkFileFilter * ff;
	int ret;

	dlg = gtk_file_chooser_dialog_new(_("Choose an icon"),
	                                  NULL,
	                                  GTK_FILE_CHOOSER_ACTION_OPEN,
	                                  GTK_STOCK_CANCEL,GTK_RESPONSE_CANCEL,
	                                  GTK_STOCK_OPEN,GTK_RESPONSE_ACCEPT,
	                                  NULL);

	//set icon
	ff = gtk_file_filter_new();
	gtk_file_filter_add_pixbuf_formats (ff);
	gtk_file_filter_set_name (ff,_("Images"));
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(dlg),ff);

	//set preview widget
	GtkWidget *preview;
	preview = gtk_image_new ();
	gtk_file_chooser_set_preview_widget (GTK_FILE_CHOOSER(dlg), preview);
	g_signal_connect (GTK_FILE_CHOOSER(dlg), "update-preview",
	                  G_CALLBACK (update_preview_cb), preview);
	

	ret = gtk_dialog_run(GTK_DIALOG(dlg));

	if (ret == GTK_RESPONSE_ACCEPT)
	{

		gchar *filename;
    	filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dlg));

		g_free(launcher->icon_path);
		launcher->icon_path = filename;
		talika_tasklist_save_launchers (launcher->tasklist);

		//update visible state now
		talika_launcher_update_visible_state (launcher);
		g_list_foreach (launcher->tasklist->visible_tasks,
	                do_talika_task_update_visible_state, NULL);
		
		//talika_tasklist_update_lists (launcher->tasklist);
		//gtk_widget_queue_resize (GTK_WIDGET (launcher->tasklist->box));
	}

	gtk_widget_destroy (dlg);
}

static gboolean
talika_launcher_button_press_event (GtkWidget * widget,
                                    GdkEventButton * event, gpointer data)
{
	TalikaTask *launcher = data;


	if (event->button == 1)
	{
		
	}
	else if (event->button == 2)
	{

	}
	else if (event->button == 3)
	{
		talika_tooltips_hide_popup_immediatly (launcher);

		if (launcher->action_menu)
			gtk_widget_destroy (launcher->action_menu);
		launcher->action_menu = NULL;

		//create context menu
		launcher->action_menu = gtk_menu_new ();
		g_object_add_weak_pointer (G_OBJECT (launcher->action_menu),
		                           (void **) &launcher->action_menu);

		GtkWidget *menu_item = gtk_image_menu_item_new_with_mnemonic (_("Unpin"));
		GtkWidget *image = gtk_image_new_from_stock (TALIKA_STOCK_UNPIN,
		                                             GTK_ICON_SIZE_MENU);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM
		                               (menu_item), image);

		gtk_menu_shell_append (GTK_MENU_SHELL(launcher->action_menu), menu_item);
		g_signal_connect (G_OBJECT (menu_item),
		                  "activate",
		                  G_CALLBACK (talika_launcher_unpin), launcher);


		menu_item = gtk_image_menu_item_new_with_mnemonic (_("Launch"));
		image = gtk_image_new_from_stock (TALIKA_STOCK_LAUNCH,
		                                  GTK_ICON_SIZE_MENU);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM
		                               (menu_item), image);

		gtk_menu_shell_append (GTK_MENU_SHELL(launcher->action_menu), menu_item);
		g_signal_connect (G_OBJECT (menu_item),
		                  "activate",
		                  G_CALLBACK (talika_launcher_launch), launcher);


		menu_item = gtk_image_menu_item_new_with_mnemonic (_("Change icon"));
		image = gtk_image_new_from_stock (TALIKA_STOCK_CHANGE_ICON,
		                                  GTK_ICON_SIZE_MENU);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM
		                               (menu_item), image);

		gtk_menu_shell_append (GTK_MENU_SHELL(launcher->action_menu), menu_item);
		g_signal_connect (G_OBJECT (menu_item),
		                  "activate",
		                  G_CALLBACK (talika_launcher_change_icon), launcher);
		

		//add attach to entry if class is nil
		if (strcmp(launcher->class_name, "nil") == 0)
		{
			//append non attached tasks
			//using hash table to disallow duplicates
			GHashTable *hash;
			hash = g_hash_table_new (g_str_hash ,g_str_equal);
			
			GList *l;
			for (l = launcher->tasklist->visible_tasks; l; l = l->next)
			{
				TalikaTask *task = l->data;

				if(task->type!=TALIKA_TASK_LAUNCHER && task->launcher == NULL)
				{
					g_hash_table_insert (hash,task->class_group_name,NULL);
				}
			}

			if (g_hash_table_size (hash)!=0)
			{

				GtkWidget *menu_item;
				GtkWidget *submenu = gtk_menu_new ();

				GHashTableIter iter;
				gpointer key, value;

				g_hash_table_iter_init (&iter, hash);
				while (g_hash_table_iter_next (&iter, &key, &value)) 
				{
					menu_item = 
						gtk_image_menu_item_new_with_label ((char*)key);
					gtk_menu_shell_append 
						(GTK_MENU_SHELL(submenu), menu_item);

					g_signal_connect (G_OBJECT (menu_item),
		                  "activate",
		                  G_CALLBACK (talika_launcher_attach_to_task), launcher);
				}

				menu_item = 
					gtk_image_menu_item_new_with_mnemonic (_("Attach task"));
				gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu_item),submenu);
				                           
				gtk_menu_shell_append 
					(GTK_MENU_SHELL(launcher->action_menu), menu_item);
			}

			g_hash_table_unref (hash);
		}
		
		gtk_widget_show_all (launcher->action_menu);
		gtk_menu_popup (GTK_MENU (launcher->action_menu),
		                NULL, NULL,
		                talika_launcher_position_menu, launcher,
		                3, gtk_get_current_event_time ());

		return TRUE;
	}
	
	return FALSE;
}

static void 
talika_launcher_update_visible_state (TalikaTask * launcher)
{
	GdkPixbuf *pixbuf;

	/*gtk_button_set_relief (GTK_BUTTON (task->button),
	                       task->
	                       tasklist->button_relief ? GTK_RELIEF_NORMAL :
		                       GTK_RELIEF_NONE);*/

	pixbuf = talika_task_get_icon (launcher);
	
	button_widget_set_pixbuf (BUTTON_WIDGET(launcher->button),pixbuf);
	if (pixbuf)
		g_object_unref (pixbuf);

	//gtk_widget_queue_resize (GTK_WIDGET (launcher->tasklist->box));
}

void 
do_talika_launcher_update_visible_state (gpointer data, 
                                     gpointer user_data)
{
	TalikaTask *launcher = (TalikaTask *) data;
	talika_launcher_update_visible_state (launcher);
}

TalikaTask*
talika_launcher_new(TalikaTaskList *tasklist, const gchar *spec)
{
	if(!spec)
		return NULL;

	//spec is like desktop_id:class_name
	gchar **tokens = g_strsplit (spec, ":", 3);

	//ill formated specification
	if (tokens[1] == NULL)
	{
		g_strfreev (tokens);
		return NULL;
	}

	//get desktop info, if .desktop exists
	GAppInfo *app_info =
		talika_launcher_get_app_info_from_id (tokens[0]);
	if(!app_info)
	{
		g_strfreev (tokens);
		return NULL;
	}

	//create new launcher
	TalikaTask *launcher = g_new0 (TalikaTask, 1);

	launcher->tasklist = tasklist;
	launcher->type = TALIKA_TASK_LAUNCHER;
		
	//extract desktop id and class name
	launcher->desktop_id = g_strdup (tokens[0]);
	launcher->class_name = g_strdup (tokens[1]);

	if(tokens[2])
		launcher->icon_path = g_strdup (tokens[2]);
	
	g_strfreev (tokens);

	launcher->app_info = app_info;
	launcher->exclude = FALSE;

	//create widgets
	launcher->button = button_widget_new (NULL,launcher);
	g_object_ref (launcher->button); //need to remove from tasklist->box often

	//drag drop connect
	static const GtkTargetEntry targets[] = {
		{"application/x-wnck-window-id", 0, 0},
		{"text/uri-list", 0, 1}
	};

	gtk_widget_set_name (launcher->button, "tasklist-button");
	gtk_button_set_relief (GTK_BUTTON(launcher->button), GTK_RELIEF_NONE);
	
	g_object_add_weak_pointer (G_OBJECT (launcher->button),
	                           (void **) &launcher->button);

	gtk_drag_source_set (GTK_WIDGET (launcher->button),
	                     GDK_BUTTON1_MASK, targets, 1, GDK_ACTION_MOVE);
	gtk_drag_dest_set (GTK_WIDGET (launcher->button),
	                   GTK_DEST_DEFAULT_DROP, targets, 2,
	                   GDK_ACTION_MOVE);
	
	GdkPixbuf *buf = talika_task_get_icon (launcher);

	if(buf)
	{
		button_widget_set_pixbuf (BUTTON_WIDGET(launcher->button),buf);
		if (buf)
			g_object_unref (buf);
	}
	
	g_signal_connect (G_OBJECT (launcher->button),
	                  "button_press_event",
	                  G_CALLBACK (talika_launcher_button_press_event), launcher);
	g_signal_connect (G_OBJECT (launcher->button),
	                  "toggled",
	                  G_CALLBACK (talika_launcher_launch), launcher);
	
	g_signal_connect_after (launcher->button, "enter-notify-event",
	                        G_CALLBACK (talika_tooltips_enter_notify), launcher);
	g_signal_connect_after (launcher->button, "leave-notify-event",
	                        G_CALLBACK (talika_tooltips_leave_notify),
	                        launcher->tasklist);

	g_signal_connect (G_OBJECT (launcher->button), "drag_motion",
	                  G_CALLBACK (talika_task_drag_motion), launcher);


	g_signal_connect (G_OBJECT (launcher->button),
	                  "drag_data_get",
	                  G_CALLBACK (talika_task_drag_data_get), launcher);

	g_signal_connect (G_OBJECT (launcher->button),
	                  "drag_data_received",
	                  G_CALLBACK (talika_task_drag_data_received), launcher);

	g_signal_connect (G_OBJECT (launcher->button),
	                  "drag_begin",
	                  G_CALLBACK (talika_task_drag_begin), launcher);

	g_signal_connect (G_OBJECT (launcher->button), "drag_end",
	                  G_CALLBACK (talika_task_drag_end), launcher);


	g_signal_connect (G_OBJECT (launcher->button), "drag_leave",
	                  G_CALLBACK (talika_task_drag_leave), launcher);

	gtk_widget_show_all (launcher->button);
	
	return launcher;
}

void
talika_launcher_free(TalikaTask *launcher)
{
	if (!launcher)
		return;

	if (launcher->button)
	{
		g_object_remove_weak_pointer (G_OBJECT (launcher->button),
		                              (void **) &launcher->button);
		gtk_widget_destroy (launcher->button);
		launcher->button = NULL;
	}

	if (launcher->action_menu)
	{
		g_object_remove_weak_pointer (G_OBJECT (launcher->action_menu),
		                              (void **) &launcher->action_menu);
		gtk_widget_destroy (launcher->action_menu);
		launcher->action_menu = NULL;
	}

	g_free(launcher->desktop_id);
	g_free(launcher->class_name);
	g_free(launcher->icon_path);

	g_object_unref (launcher->app_info);

	launcher = NULL;
}
